#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "xvfVolume.h"
#include "vvdebugmsg.h"
#include "vvvoldesc.h"

#define LINE_SIZE       512
#define BUFFER_SIZE     512
#define FILENAME_SIZE   512

int strRev( char *str ) {
	int i, j;
	i = 0; j = strlen( str )-1;
	while( i<j ) {
		char t = str[i];
		str[i] = str[j];
		str[j] = t;
		i++; j--;
	}
	return 0;
}

char *strstrLast( char *str, char *subStr ) {
	char tmpStr[BUFFER_SIZE], tmpSubStr[BUFFER_SIZE];
	strcpy( tmpStr, str );
	strcpy( tmpSubStr, subStr );
	strRev( tmpStr );
	strRev( tmpSubStr );
	char *occurrence = strstr( tmpStr, tmpSubStr );

	if( occurrence != NULL )
		return str + strlen(occurrence) - strlen(subStr);
	return NULL;
}

template <class T>
inline T max(T a, T b, T c) { int t = (a>b)?a:b; return (t>c)?t:c; }

bool generator( char *filename, int resizeOpt, int maxLevel, FILE* batchFp, XVFVolume *volArg ) {
	Volume *vol;
	static int v = 0;

	char *extFilename, tmpFilename[FILENAME_SIZE];

	if ( strstrLast(filename, "/") == NULL )
		strcpy( tmpFilename, filename );
	else
		strcpy(tmpFilename, strstrLast(filename, "/")+1);

	if ( (extFilename = strstrLast(tmpFilename, ".xvf")) != NULL ) vol = new XVFVolume();
	else {
		vvDebugMsg::msg(1, "Error: volume file format unsupported");
		return false;
	}
	*extFilename = '\0'; extFilename++;

	// setting parameters
	vol->setVolOrganization(volArg->getVolOrganization());
	vol->setBrickSize(volArg->getBrickSize());

	if( maxLevel == -1 ) {
		vol->loadHeaders(filename);
		int maxSize = max<int>( vol->vd->vox[0], vol->vd->vox[1], vol->vd->vox[2] );
		maxLevel = 0;
		while( maxSize > vol->getBrickSize() ) { maxSize>>=1; maxLevel++; }
		vol->closeFile();
	}

	// Generate mipmaps
	float zoom = 1.0;
	char path[FILENAME_SIZE];
	getcwd(path, FILENAME_SIZE);
	for( int i = 0; i<=maxLevel; i++ ) {
		float zoom_array[3] = {zoom, zoom, zoom};

		char FNGether[FILENAME_SIZE];
		sprintf( FNGether, "%s/%s.%d", path, tmpFilename, i );
		vol->resizeByCPUAndSaveBricksToOneVolume(filename, FNGether, extFilename, zoom_array, resizeOpt);
		zoom/=2;
	}

	// Generate batch files
	bool append = true;
	if( batchFp==NULL ) {
		append = false;
		char batchFN[FILENAME_SIZE];
		sprintf(batchFN, "%s.txt", tmpFilename);
		if( (batchFp = fopen( batchFN, append?"a":"w" ))==NULL ) {
			vvDebugMsg::msg( 1, "Error: batch file open error.\n");
			return false;
		}
		fprintf( batchFp, "# dimension, # volumes\n" );
		fprintf( batchFp, "%d %d\n\n", 3, 1 );
		v = 0;
	}

	fprintf( batchFp, "# Volume %s\n", filename );
	fprintf( batchFp, "# Volume definition (volume number, brick size, maximum mipmap level, # channel\n" );
	fprintf( batchFp, "%d %d %d %d\n", v, vol->getBrickSize(), maxLevel, vol->vd->chan);
	fprintf( batchFp, "# mipmap level 0 definition (mipmap level, file name, size in X, size in Y, size in Z\n");
	zoom = 1.0;
	for( int i = 0; i<=maxLevel; i++ ) {
		fprintf(batchFp, "%d ", i);
		fprintf(batchFp, "%s/%s.%d.%s ", path, tmpFilename, i, extFilename );
		fprintf(batchFp, "%d %d %d\n", (int)(1.0f*vol->vd->vox[0]*zoom), (int)(1.0f*vol->vd->vox[1]*zoom), (int)(1.0f*vol->vd->vox[2]*zoom));

		printf("%d ", i);
		printf("%s.%d.%s ",tmpFilename, i, extFilename );
		printf("%d %d %d\n", (int)(1.0f*vol->vd->vox[0]*zoom), (int)(1.0f*vol->vd->vox[1]*zoom), (int)(1.0f*vol->vd->vox[2]*zoom));

		zoom/=2;
	}
	if(append){
		fprintf( batchFp, "\n" );
		v++;
	}
	else fclose(batchFp);

	return true;
}

bool generatorSeq( char *batchFNIn, int resizeOpt, int maxLevel, char *batchFNOut, XVFVolume *volArg ) {
        FILE *batchFpIn, *batchFpOut;
        char filename[FILENAME_SIZE], line[LINE_SIZE];
        int numVolumes;

        if( (batchFpIn = fopen( batchFNIn, "r" ))==NULL ) {
                vvDebugMsg::msg( 1, "Error: batch file open error.\n");
                return false;
        }

        // skip comment lines
        while( fgets(line, sizeof(line), batchFpIn) != NULL && (line[0] == '#' || line[0] == '\0') );
        sscanf( line, "%d", &numVolumes );

        if( batchFNOut!=NULL ) {
                if( (batchFpOut = fopen( batchFNOut, "w" ))==NULL ) {
                        vvDebugMsg::msg( 1, "Error: batch file open error.\n");
                        return false;
                }

                fprintf( batchFpOut, "# dimension, # volumes\n" );
                fprintf( batchFpOut, "%d %d\n\n", 3, numVolumes );
        } else batchFpOut = NULL;

        for( int i = 0; i<numVolumes; i++ ) {
                while( fgets(line, sizeof(line), batchFpIn) != NULL && (line[0] == '#' || line[0] == '\0') );
                sscanf( line, "%s", filename );
                generator( filename, resizeOpt, maxLevel, batchFpOut, volArg);
        }
}

bool isInt(char *str) {
	int i = 0;
	if ( str[0] == '-' ) i = 1;

	for( ; str[i]; i++ )
		if( str[i]>57 || str[i]<48 ) return false;
	return true;
}

int main(int argc, char **argv) {
	// declare an volume instance of an arbitrary volume type to store parameters 
	XVFVolume *vol = new XVFVolume();

	int i=0;
	for(i=1; i<argc; i++) {	
		// Output mipmap format -- mix or separate
		if(strcmp(argv[i], "-f") == 0) {
			i++;
			if(strcmp(argv[i], "mix") == 0)				
				vol->setVolOrganization(VOL_ORG_MIX);
			else if(strcmp(argv[i], "sep") == 0)
				vol->setVolOrganization(VOL_ORG_SEP);
		}
		// chunkSize
		else if(strcmp(argv[i], "-s") == 0) {
			i++;	
			vol->setBrickSize(atoi(argv[i]));
		}
		else {
			i--;
			break;
		}
	}
	
	if( (argc-i)<3 ) {
		printf("usage: ./MipmapGen3D [-f format -s brickSize] volumeFilename maxLevel\n");
		printf("usage: ./MipmapGen3D [-f format -s brickSize] inputBatchFilename maxLevel [outputBatchFilename]\n");
		return 1;
	}
	if( !isInt(argv[i+2]) ) {
		printf("usage: ./MipmapGen3D [-f format -s brickSize] volumeFilename maxLevel\n");
		printf("usage: ./MipmapGen3D [-f format -s brickSize] inputBatchFilename maxLevel [outputBatchFilename]\n");
		return 1;
	}

	if(argc == 4)
		generatorSeq( argv[i+1], NEAREST, atoi(argv[i+2]), argv[i+3], vol );
	else {
		if ( strstrLast(argv[i+1], ".txt") != NULL ) {
			if ( (argv[i+1]+strlen(argv[i+1]) - strstrLast(argv[i+1], ".txt"))==4 )
				generatorSeq( argv[i+1], NEAREST, atoi(argv[i+2]), NULL, vol );
		} else if ( strstrLast(argv[i+1], ".xvf") != NULL ) {
			if ( (argv[i+1]+strlen(argv[i+1]) - strstrLast(argv[i+1], ".xvf"))==4 )
				generator( argv[i+1], NEAREST, atoi(argv[i+2]), NULL, vol );
		}
	}

	return 0;
}
