#include "volume.h"
#include "vvdebugmsg.h"

#define DFT_BRICK_SIZE 128
#define FILENAME_SIZE 512

Volume::Volume() {
	vd = NULL; 

	setBrickSize(DFT_BRICK_SIZE);
	setVolOrganization(VOL_ORG_MIX);

	fpVol = NULL; 
	whereData = 0;
}

Volume::~Volume() {
	if(vd != NULL) delete vd;
	if(fpVol != NULL) fclose(fpVol);
}

void Volume::setBrickSize(int bS) {
	brickSize = bS;

	if(vd != NULL) {
		for(int i=0; i<3; i++) {
			nBricks[i] = vd->vox[i]/brickSize;
			remaingPixels[i] = vd->vox[i]%brickSize;
			if(remaingPixels[i] != 0) nBricks[i]++;
		}
		sliceBytes = vd->getSliceBytes();
		BPV = vd->getBPV();
	} else {
		nBricks[0] =  nBricks[1] =  nBricks[2] = 0;
		remaingPixels[0] = remaingPixels[1] = remaingPixels[2] = 0;
		sliceBytes = 0;
		BPV = 0;
	}
}

int Volume::getBrickSize() {
	return brickSize;
}

void Volume::setVolOrganization(unsigned int volOrg) {
	volOrganization = volOrg;
}

// Combine channels together: this function save channels of each pixel together
void Volume::resizeByCPUAndSaveBricksToOneVolume( char *filename, char *FNInit, char *FNFileType,
						 float zoom[3], int mode ) {
	switch(volOrganization) {
		case VOL_ORG_SEP:
			resizeByCPUAndSaveBricksToOneVolume_sep( filename, FNInit, FNFileType, zoom, mode );
			break;
		case VOL_ORG_MIX:
			resizeByCPUAndSaveBricksToOneVolume_mix( filename, FNInit, FNFileType, zoom, mode );
			break;
	}
}

void Volume::resizeByCPUAndSaveBricksToOneVolume_mix( char *filename, char *FNInit, char *FNFileType,
//							float zoomX, float zoomY, Image::ShrinkMode mode ) {
							float zoom[3], int mode ) {
	// read required information in the image header
	FILE* fpIn = loadHeaders( filename );

	if( fpIn == NULL ) {
		vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveBricksToOneVolume() - loadHeaders() returns NULL." );
		return ;
	}

	int shrinkSize[3] = { (int)(zoom[0]*vd->vox[0]), (int)(zoom[1]*vd->vox[1]), (int)(zoom[2]*vd->vox[2]) };
	char FNameGether[FILENAME_SIZE];
	sprintf( FNameGether, "%s.%s", FNInit, FNFileType);
	FILE* fpOut = saveAppend( NULL, FNameGether, shrinkSize, vd, 0, NULL );

	int shrinkNChunks[3] = { (int)(shrinkSize[0]/brickSize), (int)(shrinkSize[1]/brickSize), (int)(shrinkSize[2]/brickSize) };
	int shrinkNRemainingPixels[3] = { shrinkSize[0]%brickSize, shrinkSize[1]%brickSize, shrinkSize[2]%brickSize };
	if( shrinkNRemainingPixels[0] != 0 ) shrinkNChunks[0]++;
	if( shrinkNRemainingPixels[1] != 0 ) shrinkNChunks[1]++;
	if( shrinkNRemainingPixels[2] != 0 ) shrinkNChunks[2]++;

	
	float zoomReal[3] = { 1.*(vd->vox[0]-1)/(shrinkSize[0]-1), 1.*(vd->vox[1]-1)/(shrinkSize[1]-1), 1.*(vd->vox[2]-1)/(shrinkSize[2]-1) };

	int shrinkEnd[3];
	int shrinkBrickPixels[3];
	for(int col=0; col<shrinkNChunks[0]; col++) {
		shrinkEnd[0] = (col == (shrinkNChunks[0]-1) && shrinkNRemainingPixels[0] != 0)? shrinkSize[0] : (col+1)*brickSize;
		shrinkBrickPixels[0] = (col == (shrinkNChunks[0]-1) && shrinkNRemainingPixels[0] != 0)? shrinkNRemainingPixels[0] : brickSize;

		for(int row=0; row<shrinkNChunks[1]; row++) {
			shrinkEnd[1] = (row == (shrinkNChunks[1]-1) && shrinkNRemainingPixels[1] != 0 )? shrinkSize[1] : (row+1)*brickSize;
			shrinkBrickPixels[1] = (row == (shrinkNChunks[1]-1) && shrinkNRemainingPixels[1] != 0)? shrinkNRemainingPixels[1] : brickSize;			

			for(int depth=0; depth<shrinkNChunks[2]; depth++) {
				shrinkEnd[2] = (depth == (shrinkNChunks[2]-1) && shrinkNRemainingPixels[2] != 0 )? shrinkSize[2] : (depth+1)*brickSize;
				shrinkBrickPixels[2] = (depth == (shrinkNChunks[2]-1) && shrinkNRemainingPixels[2] != 0)? shrinkNRemainingPixels[2] : brickSize;

				int brickBytes = shrinkBrickPixels[0]*shrinkBrickPixels[1]*shrinkBrickPixels[2]*BPV;

				uchar *raw = new uchar[brickBytes];
				memset(raw, 0, brickBytes);
				int incRaw = 0;

				int x, y, z; x = y = z = 0;
				float fOri[3] = {0, 0, 0};
				int iOri[3] = {0, 0, 0};
				for(int shrinkZ = depth*brickSize; shrinkZ<shrinkEnd[2]; shrinkZ++, z++, y=0, x=0) {
					fOri[2] = 1.*shrinkZ*zoomReal[2];
					switch( mode ){
						case NEAREST:
							iOri[2] = round( fOri[2] );
							break;
						case BILINEAR:
							break;
					}

					for(int shrinkY = row*brickSize; shrinkY<shrinkEnd[1]; shrinkY++, y++, x=0) {
						fOri[1] = 1.*shrinkY*zoomReal[1];
						switch( mode ){
							case NEAREST:
								iOri[1] = round( fOri[1] );
								break;
							case BILINEAR:
								break;
						}

						for(int shrinkX = col*brickSize; shrinkX<shrinkEnd[0]; shrinkX++, x++) {
							fOri[0] = 1.*shrinkX*zoomReal[0];							
							switch( mode ){
							case NEAREST:
								iOri[0] = round( fOri[0] );
								break;
							case BILINEAR:
								break;
							}
							if ( fseek( fpIn, whereData + (iOri[2]*sliceBytes + BPV*(iOri[1]*vd->vox[0] + iOri[0])) , SEEK_SET ) != 0 ) {
								vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - seek origin volume.");
								vvDebugMsg::msg( 1, "                                                     - start point: ", whereData + (iOri[2]*sliceBytes + BPV*(iOri[1]*vd->vox[0] + iOri[0])));
								fclose(fpIn);
								fclose (fpOut);
								return ;
							}

							if ( fread( raw + incRaw, BPV, 1, fpIn ) != 1 ) {
								vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - read origin volume.");
								fclose(fpIn);
								fclose (fpOut);
								return ;
							}
							incRaw += BPV;
						}
					}
				}

				if ( (fpOut = saveAppend( fpOut, NULL, shrinkBrickPixels, NULL, BPV, raw )) == NULL ) {
					vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - save chunked volume: ", FNameGether);
					fclose (fpIn);
					fclose (fpOut);
					return ;
				}

/*				vvVolDesc *vdOut = new vvVolDesc();
				*vdOut = *vd;
				vdOut->vox[0] = shrinkBrickPixels[0];
				vdOut->vox[1] = shrinkBrickPixels[1];
				vdOut->vox[2] = shrinkBrickPixels[2];
				//vdOut->removeSequence();
				vdOut->addFrame(raw, vvVolDesc::ARRAY_DELETE);
		
				char FNInitGather[ FILENAME_SIZE ];
				sprintf( FNInitGather, "%s_%d_%d_%d.%s", FNInit, col, row, depth, "xvf" );
				printf("File: %s ", FNInitGather);
				if ( save(FNInitGather, vdOut) != true ) {
					vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - save chunked volume: ", FNInitGather);
					fclose (fpIn);
					return ;
				}
*/
			}
		}
	}

	fclose(fpIn);
	fclose (fpOut);
}							

void Volume::resizeByCPUAndSaveBricksToOneVolume_sep( char *filename, char *FNInit, char *FNFileType,
//							float zoomX, float zoomY, Image::ShrinkMode mode ) {
							float zoom[3], int mode ) {
	// read required information in the image header
	FILE* fpIn = loadHeaders( filename );

	if( fpIn == NULL ) {
		vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveBricksToOneVolume() - loadHeaders() returns NULL." );
		return ;
	}

	int shrinkSize[3] = { (int)(zoom[0]*vd->vox[0]), (int)(zoom[1]*vd->vox[1]), (int)(zoom[2]*vd->vox[2]) };
	char FNameGether[FILENAME_SIZE];
	sprintf( FNameGether, "%s.%s", FNInit, FNFileType);
	FILE* fpOut = saveAppend( NULL, FNameGether, shrinkSize, vd, 0, NULL );

	int shrinkNChunks[3] = { (int)(shrinkSize[0]/brickSize), (int)(shrinkSize[1]/brickSize), (int)(shrinkSize[2]/brickSize) };
	int shrinkNRemainingPixels[3] = { shrinkSize[0]%brickSize, shrinkSize[1]%brickSize, shrinkSize[2]%brickSize };
	if( shrinkNRemainingPixels[0] != 0 ) shrinkNChunks[0]++;
	if( shrinkNRemainingPixels[1] != 0 ) shrinkNChunks[1]++;
	if( shrinkNRemainingPixels[2] != 0 ) shrinkNChunks[2]++;

	
	float zoomReal[3] = { 1.*(vd->vox[0]-1)/(shrinkSize[0]-1), 1.*(vd->vox[1]-1)/(shrinkSize[1]-1), 1.*(vd->vox[2]-1)/(shrinkSize[2]-1) };

	int shrinkEnd[3];
	int shrinkBrickPixels[3];
	for(int col=0; col<shrinkNChunks[0]; col++) {
		shrinkEnd[0] = (col == (shrinkNChunks[0]-1) && shrinkNRemainingPixels[0] != 0)? shrinkSize[0] : (col+1)*brickSize;
		shrinkBrickPixels[0] = (col == (shrinkNChunks[0]-1) && shrinkNRemainingPixels[0] != 0)? shrinkNRemainingPixels[0] : brickSize;

		for(int row=0; row<shrinkNChunks[1]; row++) {
			shrinkEnd[1] = (row == (shrinkNChunks[1]-1) && shrinkNRemainingPixels[1] != 0 )? shrinkSize[1] : (row+1)*brickSize;
			shrinkBrickPixels[1] = (row == (shrinkNChunks[1]-1) && shrinkNRemainingPixels[1] != 0)? shrinkNRemainingPixels[1] : brickSize;			

			for(int depth=0; depth<shrinkNChunks[2]; depth++) {
				shrinkEnd[2] = (depth == (shrinkNChunks[2]-1) && shrinkNRemainingPixels[2] != 0 )? shrinkSize[2] : (depth+1)*brickSize;
				shrinkBrickPixels[2] = (depth == (shrinkNChunks[2]-1) && shrinkNRemainingPixels[2] != 0)? shrinkNRemainingPixels[2] : brickSize;

				int brickBytes = shrinkBrickPixels[0]*shrinkBrickPixels[1]*shrinkBrickPixels[2]*vd->bpc;

				uchar *raw[vd->chan];
				int incRaw = 0;
				for( int i=0; i<vd->chan; i++ ) {
					raw[i] = NULL;
					raw[i] = new unsigned char [brickBytes];
					memset(raw[i], 0, brickBytes);
				};

				int x, y, z; x = y = z = 0;
				float fOri[3] = {0, 0, 0};
				int iOri[3] = {0, 0, 0};
				for(int shrinkZ = depth*brickSize; shrinkZ<shrinkEnd[2]; shrinkZ++, z++, y=0, x=0) {
					fOri[2] = 1.*shrinkZ*zoomReal[2];
					switch( mode ){
						case NEAREST:
							iOri[2] = round( fOri[2] );
							break;
						case BILINEAR:
							break;
					}

					for(int shrinkY = row*brickSize; shrinkY<shrinkEnd[1]; shrinkY++, y++, x=0) {
						fOri[1] = 1.*shrinkY*zoomReal[1];
						switch( mode ){
							case NEAREST:
								iOri[1] = round( fOri[1] );
								break;
							case BILINEAR:
								break;
						}

						for(int shrinkX = col*brickSize; shrinkX<shrinkEnd[0]; shrinkX++, x++) {
							fOri[0] = 1.*shrinkX*zoomReal[0];							
							switch( mode ){
							case NEAREST:
								iOri[0] = round( fOri[0] );
								break;
							case BILINEAR:
								break;
							}
							if ( fseek( fpIn, whereData + (iOri[2]*sliceBytes + BPV*(iOri[1]*vd->vox[0] + iOri[0])) , SEEK_SET ) != 0 ) {
								vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - seek origin volume.");
								vvDebugMsg::msg( 1, "                                                     - start point: ", whereData + (iOri[2]*sliceBytes + BPV*(iOri[1]*vd->vox[0] + iOri[0])));
								fclose(fpIn);
								fclose (fpOut);
								return ;
							}

							for( int i=0; i<vd->chan; i++ )
								if ( fread( raw[i] + incRaw, vd->bpc, 1, fpIn ) != 1 ) {
									vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - read origin volume.");
									fclose(fpIn);
									fclose(fpOut);
									return ;
								}

							incRaw += vd->bpc;
						}
					}
				}

				for(int i=0; i<vd->chan; i++)
					if ( (fpOut = saveAppend( fpOut, NULL, shrinkBrickPixels, NULL, vd->bpc, raw[i] )) == NULL ) {
						vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - save chunked volume: ", FNameGether);
						fclose (fpIn);
						fclose(fpOut);
						return ;
					}
				for(int i=0; i<vd->chan; i++) delete [] raw[i];

/*				vvVolDesc *vdOut = new vvVolDesc();
				*vdOut = *vd;
				vdOut->vox[0] = shrinkBrickPixels[0];
				vdOut->vox[1] = shrinkBrickPixels[1];
				vdOut->vox[2] = shrinkBrickPixels[2];
				//vdOut->removeSequence();
				vdOut->addFrame(raw, vvVolDesc::ARRAY_DELETE);
		
				char FNInitGather[ FILENAME_SIZE ];
				sprintf( FNInitGather, "%s_%d_%d_%d.%s", FNInit, col, row, depth, "xvf" );
				printf("File: %s ", FNInitGather);
				if ( save(FNInitGather, vdOut) != true ) {
					vvDebugMsg::msg( 1, "Error: Volume::resizeByCPUAndSaveChunksToOneVolume() - save chunked volume: ", FNInitGather);
					fclose (fpIn);
					return ;
				}
*/
			}
		}
	}

	fclose(fpIn);
	fclose (fpOut);
}							
