#include "xvfVolume.h"
#include "vvtokenizer.h"
#include "vvdebugmsg.h"
#include "vvfileio.h"

XVFVolume::XVFVolume() {
	vd = new vvVolDesc();
	//vd -> setDefaults();
}

XVFVolume::XVFVolume(char* filename)
{
	vd = new vvVolDesc();

	if((fpVol = loadHeaders(filename)) == NULL) {
		vvDebugMsg::msg(1, "Error: cannot open file: ", filename);
		return;
	}
	whereData = ftell(fpVol);
}


XVFVolume::~XVFVolume() {
}

bool XVFVolume::load(char *filename, vvVolDesc *vdIn) {
	if(filename == NULL) {
		vvDebugMsg::msg(1, "Empty filename!");
		return false;
	}

	vvFileIO *fio =  new vvFileIO();

	vdIn->setFilename(filename);
	//if(fio->loadXVFFile(vdIn) != vvFileIO::OK) {
	if(fio->loadVolumeData(vdIn)  != vvFileIO::OK) { 
		vvDebugMsg::msg(1, "loadXVFFile fail!");
		return false;
	}

	delete fio;

	setBrickSize(brickSize);
	return true;
}

bool XVFVolume::save(char *filename, vvVolDesc *vdOut) {
	if(filename == NULL) {
		vvDebugMsg::msg(1, "Empty filename!");
		return false;
	}

	vvFileIO *fio =  new vvFileIO();

	//if(fio->saveXVFFile(vdOut) != vvFileIO::OK) {
	vdOut->setFilename(filename); 
	fio->setCompression(false);
	if(fio->saveVolumeData(vdOut, true) != vvFileIO::OK) {
		vvDebugMsg::msg(1, "saveXVFFile fail!");
		return false;
	}
	delete fio;
	return true;
}

FILE * XVFVolume::loadHeaders(char *filename) {
	const char xvfID[4] = "XVF";
	FILE* fp;                                       // volume file pointer
	vvTokenizer* tok;
	vvTokenizer::TokenType ttype;                   // currently processed token type
	int i;                                          // counters
	int frameSize;                                  // size of a frame in bytes
	bool done;
	int encodedSize;                                // size of encoded data array
	vvToolshed::EndianType endian;

	vvDebugMsg::msg(1, "vvFileIO::loadXVFFile()");

	vd->setFilename(filename);
	if (vd->getFilename()==NULL) return NULL;

	if ( (fp = fopen(vd->getFilename(), "rb")) == NULL) 
	{
		vvDebugMsg::msg(1, "Error: Cannot open file.");
		return NULL;
	}

	// Read magic code:
	for (i=0; i<int(strlen(xvfID)); ++i)
	{
		if (fgetc(fp) != xvfID[i])
		{
			fclose(fp);
			// vvDebugMsg::msg(1, "Trying to load old style XVF format.");
			// return loadXVFFileOld(vd);
			vvDebugMsg::msg(1, "Old style XVF format unsupported.");
			return NULL;
		}
	}
	fseek(fp, 0, SEEK_SET);

	vd->removeSequence();                           // delete previous volume sequence

	// Read header:
	tok = new vvTokenizer(fp);
	tok->setCommentCharacter('#');
	tok->setEOLisSignificant(false);
	tok->setCaseConversion(vvTokenizer::VV_UPPER);
	tok->setParseNumbers(true);
	tok->setWhitespaceCharacter(' ');

	done = false;
	while (!done)
	{
		  // Read a token:
		ttype = tok->nextToken();
		if (ttype != vvTokenizer::VV_WORD)
		{
			cerr << "Error in line " << tok->getLineNumber() << " of XVF file" << endl;
			tok->nextLine();
		}
		else
		{
			if (strcmp(tok->sval, "XVF")==0)
			{
				tok->nextLine();
			}
			else if (strcmp(tok->sval, "VERSION")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				cerr << "Reading XVF file version " << tok->nval << endl;
			}
			else if (strcmp(tok->sval, "VOXELS")==0)
			{
				for (i=0; i<3; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->vox[i] = int(tok->nval);
				}
			}
			else if (strcmp(tok->sval, "TIMESTEPS")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->frames = int(tok->nval);
			}
			else if (strcmp(tok->sval, "BPC")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->bpc = int(tok->nval);
			}
			else if (strcmp(tok->sval, "CHANNELS")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->chan = int(tok->nval);
			}
			else if (strcmp(tok->sval, "DIST")==0)
			{
				for (i=0; i<3; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->dist[i] = tok->nval;
				}
			}
			else if (strcmp(tok->sval, "ENDIAN")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_WORD);
				if (strcmp(tok->sval, "LITTLE")==0) endian = vvToolshed::VV_LITTLE_END;
				else endian = vvToolshed::VV_BIG_END;
			}
			else if (strcmp(tok->sval, "DTIME")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->dt = tok->nval;
			}
			else if (strcmp(tok->sval, "MINMAX")==0)
			{
				for (i=0; i<2; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->real[i] = tok->nval;
				}
			}
			else if (strcmp(tok->sval, "POS")==0)
			{
				for (i=0; i<3; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->pos[i] = tok->nval;
				}
			}
			else if (strcmp(tok->sval, "CHANNELNAMES")==0)
			{
				if (vd->chan<1) tok->nextLine();
				else
				{
					for (i=0; i<vd->chan; ++i)
					{
						ttype = tok->nextToken();
						assert(ttype == vvTokenizer::VV_WORD);
						vd->setChannelName(i, tok->sval);
					}
				}
			}
			else if (strcmp(tok->sval, "TF_PYRAMID")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFPyramid(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "TF_BELL")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFBell(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "TF_COLOR")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFColor(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "TF_CUSTOM")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFCustom(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "ICON")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->iconSize = int(tok->nval);
				tok->nextLine();

				// Read icon:
				if (vd->iconSize>0)
				{
					delete[] vd->iconData;
					int iconBytes = vd->iconSize * vd->iconSize * vvVolDesc::ICON_BPP;
					vd->iconData = new uchar[iconBytes];
					fseek(fp, tok->getFilePos(), SEEK_SET);
					int encodedSize = vvToolshed::read32(fp);
					if (encodedSize>0)                      // compressed icon?
					{
						uchar* encoded = new uchar[encodedSize];
						if ((int)fread(encoded, 1, encodedSize, fp) != encodedSize)
						{
							cerr << "Error: Insuffient compressed icon data in file." << endl;
							fclose(fp);
							delete[] vd->iconData;
							vd->iconData = NULL;
							delete[] encoded;
							return NULL;
						}
						if (vvToolshed::decodeRLE(vd->iconData, encoded, encodedSize, vvVolDesc::ICON_BPP, iconBytes) < 0)
						{
							cerr << "Error: Decoding exceeds icon size." << endl;
							fclose(fp);
							delete[] vd->iconData;
							vd->iconData = NULL;
							delete[] encoded;
							return NULL;
						}
						delete[] encoded;
					}
					else                                    // uncompressed icon
					{
						if ((int)fread(vd->iconData, 1, iconBytes, fp) != iconBytes)
						{
							cerr << "Error: Insuffient uncompressed icon data in file." << endl;
							fclose(fp);
							delete[] vd->iconData;
							vd->iconData = NULL;
							return NULL;
						}
					}
					tok->setFilePos(fp);
				}
			}
			else if (strcmp(tok->sval, "VOL_ORG")==0)
			{
				tok->nextToken();
				setVolOrganization((int)(tok->nval));
			}
			else if (strcmp(tok->sval, "VOXELDATA")==0)
			{
				tok->nextLine();
				done = true;
			}
			else
			{
				if (ttype == vvTokenizer::VV_WORD)
				{
					cerr << "Ignoring unknown header entry: " << tok->sval << endl;
				}
				tok->nextLine();
			}
		}
	}

	frameSize = vd->getFrameBytes();

//	if ((_sections & RAW_DATA) != 0) {
	fseek(fp, tok->getFilePos(), SEEK_SET);
	encodedSize = vvToolshed::read32(fp);
//	}

	if(encodedSize>0) {
		vvDebugMsg::msg(1, "Encoded volume file unsupported!");
		fclose(fp);
		return NULL;
	}

	// Load volume data:
/*	if ((_sections & RAW_DATA) != 0)
	{
		fseek(fp, tok->getFilePos(), SEEK_SET);
		encoded = new uchar[frameSize];
		for (f=0; f<vd->frames; ++f)
		{
			raw = new uchar[frameSize];                 // create new data space for volume data
			encodedSize = vvToolshed::read32(fp);
			if (encodedSize>0)
			{
				if ((int)fread(encoded, 1, encodedSize, fp) != encodedSize)
				{
					vvDebugMsg::msg(1, "Error: Insuffient voxel data in file.");
					fclose(fp);
					delete[] raw;
					delete[] encoded;
					return NULL;
				}
				if (vvToolshed::decodeRLE(raw, encoded, encodedSize, vd->getBPV(), frameSize) < 0)
				{
					vvDebugMsg::msg(1, "Error: Decoding exceeds frame size.");
					fclose(fp);
					delete[] raw;
					delete[] encoded;
					return NULL;
				}
			}
			else                                        // no encoding
			{
				if ((int)fread(raw, 1, frameSize, fp) != frameSize)
				{
					vvDebugMsg::msg(1, "Error: Insuffient voxel data in file.");
					fclose(fp);
					delete[] raw;
					delete[] encoded;
					return NULL;
				}
			}
			vd->addFrame(raw, vvVolDesc::ARRAY_DELETE);
		}
		delete[] encoded;
	}
*/
	setBrickSize(brickSize);

	whereData = ftell(fp);
	fpVol = fp;
	return fp;
}

FILE* XVFVolume::saveAppend (FILE *fp, char *filename, int volSize[3], vvVolDesc *vdOut, int BPV, uchar *raw) {
	uchar* encoded = NULL;                          // encoded volume data
	int i;                                          // counters
	int frames;                                     // volume animation frames
	int frameSize;                                  // frame size
	int numTF;                                      // number of transfer functions to save
	int encodedSize;                                // number of bytes in encoded array

	vvDebugMsg::msg(1, "vvFileIO::saveXVFFile()");

	if( fp==NULL && vdOut!= NULL && raw==NULL )  {
		// Prepare variables:
		frames = vdOut->frames;
		if(frames>1)
			vvDebugMsg::msg(1, "Warning: XVFVolume::saveAppend() - multiple frames unsupported.");

		if (frames==0) return NULL;

		// Open file:
		// now open file to write
		if ( (fp = fopen(filename, "wb")) == NULL)
		{
			vvDebugMsg::msg(1, "Error: XVFVolume::saveAppend() - cannot open file to write.");
			return NULL;
		}

		// Force icon to be present:
		if (vdOut->iconSize==0) vdOut->makeIcon(vvVolDesc::DEFAULT_ICON_SIZE);

		// Write header:
		fprintf(fp, "XVF\n");
		fprintf(fp, "VERSION %2.1f\n", 2.0f);
		fprintf(fp, "VOXELS %d %d %d\n", volSize[0], volSize[1], volSize[2]);
		fprintf(fp, "TIMESTEPS %d\n", vdOut->frames);
		fprintf(fp, "BPC %d\n", vdOut->bpc);
		fprintf(fp, "CHANNELS %d\n", vdOut->chan);
		fprintf(fp, "DIST %g %g %g\n", vdOut->dist[0], vdOut->dist[1], vdOut->dist[2]);
		fprintf(fp, "ENDIAN %s\n", (vvToolshed::getEndianness()==vvToolshed::VV_LITTLE_END) ? "LITTLE" : "BIG");
		fprintf(fp, "DTIME %g\n", vdOut->dt);
		fprintf(fp, "MINMAX %g %g\n", vdOut->real[0], vdOut->real[1]);
		fprintf(fp, "POS %g %g %g\n", vdOut->pos[0], vdOut->pos[1], vdOut->pos[2]);
		fprintf(fp, "VOL_ORG %d\n", getVolOrganization());

		// Write channel names:
		fprintf(fp, "CHANNELNAMES");
		for (i=0; i<vdOut->chan; ++i)
		{
			if (vdOut->getChannelName(i)==NULL) fprintf(fp, " UNNAMED");
			else fprintf(fp, " %s", vdOut->getChannelName(i));
		}
		fprintf(fp, "\n");

		// Write transfer function widgets:
		numTF = vdOut->tf._widgets.count();
		vdOut->tf._widgets.first();
		for (i=0; i<numTF; ++i)
		{
			vdOut->tf._widgets.getData()->write(fp);
			vdOut->tf._widgets.next();
		}

		// Write icon:
		fprintf(fp, "ICON %d %d\n", vdOut->iconSize, vdOut->iconSize);
		if (vdOut->iconSize>0)
		{
			int iconBytes = vdOut->iconSize * vdOut->iconSize * vvVolDesc::ICON_BPP;
			uchar* encodedIcon = new uchar[iconBytes];
			encodedSize = vvToolshed::encodeRLE(encodedIcon, vdOut->iconData, iconBytes, vvVolDesc::ICON_BPP, iconBytes);
			if (encodedSize>=0)                           // compression possible?
			{
				vvToolshed::write32(fp, encodedSize);       // write length of encoded icon
				if ((int)fwrite(encodedIcon, 1, encodedSize, fp) != encodedSize)
				{
					cerr << "Error: Cannot write compressed icon data to file." << endl;
					fclose(fp);
					delete[] encodedIcon;
					return NULL;
				}
			}
			else
			{
				vvToolshed::write32(fp, 0);                 // write zero to indicate unencoded icon
				if ((int)fwrite(vdOut->iconData, 1, iconBytes, fp) != iconBytes)
				{
					cerr << "Error: Cannot write uncompressed icon data to file." << endl;
					fclose(fp);
					delete[] encodedIcon;
					return NULL;
				}
			}
			delete[] encodedIcon;
		}
		
		// Write volume data
		fprintf(fp, "VOXELDATA\n");
		vvToolshed::write32(fp, 0);

		return fp;
	} else if (fp!=NULL) {
		// Write volume data frame by frame:	only the first frame is saved
		if (raw==NULL)
		{
			vvDebugMsg::msg(1, "Error: XVFVolume::saveAppend() - no data available for frame");
			fclose(fp);
			delete[] encoded;
			return NULL;
		}
		frameSize = volSize[0]*volSize[1]*volSize[2]*BPV;

		if ((int)fwrite(raw, 1, frameSize, fp) != frameSize)
		{
			cerr << "Error: Cannot write voxel data to file." << endl;
			fclose(fp);
			return NULL;
		}
		return fp;
	}
	return NULL;
}
